﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
namespace CarFire
{
    /// <summary>
    /// Flag options to define attributes for individual tiles.
    /// </summary>
    [Flags]
    public enum TileFlags
    {
        Open = 0x01,
        Closed = 0x02,
        Floor = 0x04,
        Wall = 0x08,
        Default = 0x05
    }


    /// <summary>
    /// Small wrapper around a texture to provide easy access to
    /// tile rectangles.
    /// </summary>
    public class Tilemap
    {
        #region Public Properties

        /// <summary>
        /// Get the texture for this tilemap.
        /// </summary>
        public Texture2D Texture { get { return mTexture; } }

        #endregion


        #region Public Methods

        /// <summary>
        /// Construct a tilemap with a texture and dimensions in
        /// tiles.
        /// </summary>
        /// <param name="texture">The texture.</param>
        /// <param name="width">Number of tiles across.</param>
        /// <param name="height">Number of tiles down.</param>
        public Tilemap(Texture2D texture, int width, int height)
        {
            mTexture = texture;
            mWidth = width;
            mHeight = height;
            mTileW = mTexture.Width / mWidth;
            mTileH = mTexture.Height / mHeight;
        }


        /// <summary>
        /// Get a tile rectangle from a tile coordinate.
        /// </summary>
        /// <param name="point">Tile coordinates; [0,0] being the
        /// top-left tile.</param>
        /// <returns>Rectangle surrounding the tile.</returns>
        public Rectangle GetRectangleForTile(Point point)
        {
            return GetRectangleForTile(point.X, point.Y);
        }

        /// <summary>
        /// Get a tile rectangle from a tile coordinate
        /// </summary>
        /// <param name="x">X-coordinate.</param>
        /// <param name="y">Y-coordinate.</param>
        /// <returns>Rectangle surrounding the tile.</returns>
        public Rectangle GetRectangleForTile(int x, int y)
        {
            Debug.Assert(0 <= x && x < mWidth && 0 <= y && y < mHeight);
            return new Rectangle(x * mTileW, y * mTileH, mTileW, mTileH);
        }

        /// <summary>
        /// Get a tile rectangle from a tile character.
        /// </summary>
        /// <param name="tile">Tile character.</param>
        /// <returns>Rectangle surrounding the tile.</returns>
        public Rectangle GetRectangleForTile(char tile)
        {
            return mTiles[tile];
        }


        /// <summary>
        /// Get the flags associated with a tile character.
        /// </summary>
        /// <param name="tile">Tile character.</param>
        /// <returns>Tile flags.</returns>
        public TileFlags GetTileFlags(char tile)
        {
            return mFlags[tile];
        }


        /// <summary>
        /// Associate a tile character with tile coordinates.  This
        /// lets you access tiles by character.
        /// </summary>
        /// <param name="tile">Tile character.</param>
        /// <param name="point">Coordinates.</param>
        public void SetTile(char tile, Point point, TileFlags flags)
        {
            mTiles.Add(tile, GetRectangleForTile(point));
            mFlags.Add(tile, flags);
        }


        /// <summary>
        /// Draw a tile to the screen.
        /// </summary>
        /// <param name="spriteBatch">The b2bomber.</param>
        /// <param name="tile">The tile.</param>
        /// <param name="screenRect">The screen rectangle to draw at.</param>
        public void Draw(SpriteBatch spriteBatch, char tile, Rectangle screenRect)
        {
            spriteBatch.Draw(mTexture, screenRect, GetRectangleForTile(tile), Color.White);
        }

        #endregion


        #region Private Variables

        Texture2D mTexture;
        int mWidth;
        int mHeight;
        int mTileW;
        int mTileH;
        Dictionary<char, Rectangle> mTiles = new Dictionary<char, Rectangle>();
        Dictionary<char, TileFlags> mFlags = new Dictionary<char, TileFlags>();

        #endregion
    }
}
